"""
This package holds the common libraries used by all Python parts of the Optimal BPM system.
It includes messaging, queue management, database initialization, decorators, and logging
"""
__author__ = 'Nicklas Borjesson'



"""
    The permission module handles permissions for nodes.
"""


from of.broker.lib.schema_mongodb import mbe_object_id

__author__ = 'nibo'


def filter_by_group(_nodes, _permission, _user, _database_access, _error_prefix_if_not_allowed=None,
                    _use_object_id=False):
    """This function filters a list of nodes based on what groups a user belong to and its right

    :param _nodes: A list of node
    :param _permission: The permission to check. For example "canRead" or "canWrite".
    :param _user: The user to check
    :param  _database_access: A database access instance.
    :param _error_prefix_if_not_allowed: If set, the supplied error message is raised. Normally used when testing for\
        write access.
    :param _use_object_id: If true, the node has ObjectId instances and not strings.
    :return: A list of nodes any of the users' groups have the permission defined in _permission for

    """
    _result = []
    _groups = _user["groups"]
    # Loop nodes
    for _curr_row in _nodes:
        _has_permission = False
        # Check if any of the group the user belongs to..
        for _curr_group in _groups:
            # ..is among those the node has listed having the requested permission
            if _use_object_id:
                if mbe_object_id(_curr_group) in _curr_row[_permission]:
                    _has_permission = True
                    break
            elif _curr_group in _curr_row[_permission]:
                _has_permission = True
                break
        if _has_permission:
            _result.append(_curr_row)
        elif _error_prefix_if_not_allowed:
            # Raise an error if a user lacks the _permission for.
            _error = _error_prefix_if_not_allowed + ":" + _user["name"] + " doesn't have the " + _permission + \
                " for the " + _curr_row["name"] + "(" + str(_curr_row["_id"]) + ")-node."

            _database_access.logging.log_security("permission",
                                                  _error,
                                                  _user["_id"], str(_curr_row["_id"]))
            raise PermissionError(_error)

    return _result



"""
Created on Nov 6, 2012

@author: Nicklas Boerjesson
@note: This decorator raises an NotImplemented error if a certain method is decorated
"""
import os
import sys
import platform

import signal
from decorator import decorator


# TODO: BPMSettings should probably not be used by QAL(use only resources)optimalbpm.common.* (PROD-94)
from of.common.logging import write_to_log, EC_SERVICE, SEV_INFO

stop_handler = None

def _not_implemented(func, *args, **kwargs):
    raise NotImplementedError('Internal error in "' + func.__name__ + '": Not implemented.')


def not_implemented(f):
    """
    Raises a NotImplementedError error. Example: "Internal error in function_name: Not implemented."
    """
    return decorator(_not_implemented, f)


def node_id_with_env(_node_id, _environment_id):
    """
    Converts a general Optimal BPM node_id strings to a local one by XOR;ing the MongoDB objectId:s.
    :param _node_id: A node object id
    :param _environment_id: The Optimal BPM environment_id
    :return: The locally usable value
    """
    result = int(_node_id, 16) ^ int(_environment_id, 16)  # convert to integers and xor them together
    return '{:x}'.format(result)


def resolve_config_path():
    """
    Load settings from settings file, first environment variable OPTIMAL_BPM_CFG, then default locations:
    * Windows - %APPDATA%\optimalframework
    * Linux/OSX - ~/optimalframework

    :return: a qal BPM settings instance
    """

    # TODO: Break out into utils, this exists in broker as well.(ORG-112)

    # TODO: Write settings documentation.(ORG-112)
    if platform.system().lower() in ["linux", "darwin"]:
        # Linux/OSX
        _default_settings_folder = os.path.join(os.path.expanduser("~"), "optimalframework")
    elif platform.system().lower() == "windows":
        # Windows
        _default_settings_folder = os.path.join(os.getenv("APPDATA"), "optimalframework")
    else:
        raise Exception("Unsupported platform: " + platform.system().lower())

    return os.path.expanduser(os.getenv("OPTIMAL_FW_CFG", os.path.join(_default_settings_folder, "config.json")))



def signal_handler_unix(_signal, _frame):
    """
    Manage system signals
    :param _signal: The Posix signal See: http://pubs.opengroup.org/onlinepubs/009695399/
    :param _frame: Unused, interrupted stack frame.
    """
    _reason = None
    if _signal == signal.SIGHUP:
        _reason = 'Got SIGTERM, restarting..'
        stop_handler(_reason, _restart=True)
    else:
        if _signal == signal.SIGQUIT:
            _reason = 'Ctrl+D pressed, shutting down..'
        elif _signal == signal.SIGINT:
            _reason = 'Ctrl+C pressed, shutting down..'
        elif _signal == signal.SIGTERM:
            _reason = 'Got SIGTERM, shutting down..'

        stop_handler(_reason)

def signal_handler_windows():
    write_to_log("Windows signal handler called", _category=EC_SERVICE, _severity=SEV_INFO)
    stop_handler('The process has been terminated, shutting down..')

def register_signals(_stop_handler):
    """
    Register a callback to gracefully handle the system being externally shut down (terminated)
    :param _stop_handler: A callback that helps the system shut down gracefully.
    """

    if _stop_handler:
        global stop_handler
        stop_handler = _stop_handler
    else:
        raise Exception("No stop handler, probably an internal error, needed for graceful shut down.")
    if os.name == "nt":
        try:
            import win32api
        except ImportError:
            version = ".".join(map(str, sys.version_info[:2]))
            raise Exception("pywin32 not installed for Python " + version)
        else:
            win32api.SetConsoleCtrlHandler(signal_handler_windows, True)
            write_to_log("Registered win32 ctrl handler", _category=EC_SERVICE, _severity=SEV_INFO)
    else:
        signal.signal(signal.SIGINT, signal_handler_unix)
        signal.signal(signal.SIGTERM, signal_handler_unix)
        signal.signal(signal.SIGQUIT, signal_handler_unix)
        signal.signal(signal.SIGHUP, signal_handler_unix)


def make_log_prefix(_address):
    """
    Create the log prefix for a broker, used in log texts.
    """
    return str(os.getpid()) + "-" + _address + ":"


"""
The monitor module implements the Monitor and MonitorThread classes.
"""
import os
import traceback
import threading
from queue import Empty, Queue

from of.common.logging import write_to_log, EC_UNCATEGORIZED, SEV_DEBUG, SEV_ERROR, EC_INTERNAL
from of.schemas.constants import zero_object_id

__author__ = 'Nicklas Boerjesson'


class MonitorThread(threading.Thread):
    #: If true, the thread is terminated
    terminated = None

    def __init__(self, _monitor):
        self.terminated = False
        super(MonitorThread, self).__init__(target=_monitor.monitor,
                                            name=_monitor.__class__.__name__ + "_queue monitor thread")


class Monitor(object):
    """
    The monitor class monitors a given queue for new items and calls a handler when new items appear on the queue.
    """

    #: The queue to be monitored
    queue = None
    #: The handler that reacts on each queue item
    handler = None
    #: The Optimal Framework process id the monitor is running within
    process_id = None
    #: Called each time the queue is polled
    before_get_queue = None
    #: A prefix that is applied to all log items
    log_prefix = None
    #: The thread that monitors the queue
    monitor_thread = None

    def __init__(self, _handler, _queue=None):

        self.log_prefix = self.__class__.__name__ + "(" + str(
            _handler.__class__.__name__) + "): "

        if _queue:
            self.queue = _queue
        else:
            self.queue = Queue()

        self.handler = _handler
        self.handler.on_monitor_init(self)
        if self.queue:
            self.start()

    def write_dbg_info(self, _data):
        write_to_log(self.log_prefix + _data, _category=EC_UNCATEGORIZED,
                                  _severity=SEV_DEBUG, _process_id=self.process_id)

    def after_get_queue(self):
        """
        Implement this for something to happen immidiately after a each queue check has happened.
        """
        pass

    def monitor(self):
        """
        Monitors the queue. Stops when the queue-threads' terminated attribute is set to True
        """
        self.write_dbg_info("In monitor thread monitoring queue.")
        while not self.monitor_thread.terminated:
            try:
                _item = self.queue.get(True, .1)
                try:
                    self.handler.handle(_item)
                except Exception as e:
                    write_to_log(self.log_prefix + "Error handling item:" + str(e) + "\nData:\n" + str(_item) +
                                 "\nTraceback:" + traceback.format_exc(), _category=EC_INTERNAL,
                                 _severity=SEV_ERROR)
            except Empty:
                pass
            except Exception as e:
                self.monitor_thread.terminated = True
                raise Exception(write_to_log(self.log_prefix + "Terminating, error accessing own queue:" + str(e),
                                             _category=EC_INTERNAL, _severity=SEV_ERROR))

            self.after_get_queue()

        self.write_dbg_info("Monitor thread stopped.")

    def stop(self, _user_id=zero_object_id, _reverse_order=None):
        """
        Stops the monitor and tells the handler to shut down

        :param _user_id: The Optimal BPM user Id that is stopping the monitor
        :param _reverse_order: If true, the monitor is kept running while the handler is shut down.
        """
        if not _reverse_order:
            self.write_dbg_info("Told to stop, ceasing monitoring.")
            self.monitor_thread.terminated = True

        # TODO: Handle any residual items on the queue.(PROD-28)
        # There should be a general framework for unhandled queue items

        self.write_dbg_info("Shutting down handler(" + str(self.handler.__class__.__name__) + ").")
        self.handler.shut_down(_user_id)
        self.write_dbg_info("Handler shut down.")
        if _reverse_order:
            self.write_dbg_info("Told to stop, ceasing monitoring.")
            self.monitor_thread.terminated = True

    def start(self):
        """
        Start monitoring the queue
        """

        if self.monitor_thread and not self.monitor_thread.terminated:
            raise Exception(write_to_log(self.log_prefix + "The queue monitor is already running.",
                                         _category=EC_INTERNAL, _severity=SEV_ERROR))
        self.monitor_thread = MonitorThread(_monitor=self)
        self.monitor_thread.terminated = False
        self.monitor_thread.start()
        self.write_dbg_info("Running, monitoring thread: " + str(self.monitor_thread.name))



"""
Contains functionality for reading settings from ini- and json-files

:copyright: Copyright 2010-2014 by Nicklas Boerjesson
:license: BSD, see LICENSE for details.
"""

import configparser
import os
import json

class INISettings(object):
    """This class is responsible for reading settings from ini-files and holding them in memory."""

    parser = None
    filename = None
    dirname = None
    def reload(self, _filename):
        """Reload all information"""
        self.parser.read(_filename)
        self.filename = _filename
        self.dirname = os.path.dirname(_filename)

    def handle_path(self, _path):
        if os.path.isabs(_path):
            return _path
        elif _path[0] == "~":
            return os.path.expanduser(_path)
        else:
            return os.path.join(self.dirname, _path)

    def get(self, _section, _option, _default=None):
        """Get a certain option"""
        if self.parser.has_section(_section):
            if not self.parser.has_option(_section, _option):
                return _default
            else:
                return self.parser.get(_section, _option)
        else:
            return _default


    def get_path(self, _section, _option, _default=None):
        return self.handle_path(self.get(_section, _option, _default))

    def __init__(self, _filename = None):
        """Constructor"""
        if _filename:
            self.filename = _filename
            # TODO: Add better config support.
            self.reload(_filename)


class JSONXPath(object):
    """This class is responsible for reading settings from json-files and holding them in memory."""

    filename = None
    dirname = None
    data = None
    def reload(self, _filename):
        """Reload all information"""
        with open(_filename, "r") as f:
            self.data = json.load(f)
        self.filename = _filename
        self.dirname = os.path.dirname(_filename)

    def find(self, _xpath):
        _parts = _xpath.split("/")
        _curr_node = self.data
        for _curr_part in _parts:
            if _curr_part in _curr_node:
                _curr_node = _curr_node[_curr_part]
            else:
                return False
        return _curr_node


    def handle_path(self, _path):
        if os.path.isabs(_path):
            return _path
        elif _path[0] == "~":
            return os.path.expanduser(_path)
        else:
            return os.path.join(self.dirname, _path)

    def get(self, _xpath, _default=None):
        """Get a certain option"""
        _node = self.find(_xpath)
        if _node is not None:
            return _node
        else:
            return _default


    def get_path(self, _xpath, _default=None):
        return self.handle_path(self.get(_xpath, _default))

    def __init__(self, _filename = None):
        """Constructor"""

        if _filename:
            self.filename = _filename
            # TODO: Add better config support.
            self.reload(_filename)



"""
This module contains the implementation of the WebSocketHandler class
"""
import logging
from asyncore import write
from threading import Lock

from of.common.logging import write_to_log, SEV_ERROR, EC_COMMUNICATION, EC_PROBE, EC_UNCATEGORIZED, EC_INTERNAL
from of.common.messaging.constants import PROTOCOL_ERROR, GOING_AWAY
from of.common.messaging.factory import reply_with_error_message
from of.common.queue.handler import Handler
from of.schemas.constants import schema_categories

__author__ = 'Nicklas Borjesson'


class WebSocketHandler(Handler):
    """
        Message handler adds support to the base handler for handling messages for web sockets
    """
    #: A dictionary containing the peers of the connected web sockets
    peers = None
    #: The peer adress of this web socket
    address = None
    #: A map between adresses and peers
    address__session = None
    #: An instance of the MBE schema tools
    schema_tools = None
    # Short cut between message categories and handlers
    category_shortcut = None

    def __init__(self, _process_id, _peers, _schema_tools, _address):
        """
        Initialize the handler. create Lock instance
        :param _process_id: The BPM process id of the running scripts
        :param _peers: A dict of logged in peers
        :param _schema_tools: An instance of the mbe SchemaTools class
        :param _address: The peer address of the peer
        """
        super(WebSocketHandler, self).__init__(_process_id)
        self.peers = _peers
        self.address = _address
        self.schema_tools = _schema_tools
        self.address__session = {}
        self.category_shortcut = {}
        self._last_message_id = 0
        self.web_socket_lock = Lock()


    def handle_error(self, _error, _category = EC_COMMUNICATION, _severity = SEV_ERROR,
                     _web_socket= None, _close_socket = None, _message_to_reply_to=None):
        """

        :param _error: The error message
        :param _category: The category of the error
        :param _severity: Error severity
        :param _web_socket: If the sender was external, the web socket
        :param _close_socket: Close the web socket
        :param _message_to_reply_to: The source message, if set, the error message is sent to the sender
        :return:
        """
        """ Handles and logs errors, decides whether to keep the connection open if an error occurrs


        """
        _error = write_to_log(_data=self.log_prefix + "In handler.handle_error error with error :" +
                           str(_error), _category=_category, _severity=_severity)
        if _web_socket:

            if _message_to_reply_to:
                _web_socket.send_message(reply_with_error_message(self, _message_to_reply_to, _error))

            if _close_socket:
                _web_socket.close(code=PROTOCOL_ERROR, reason=_error)

    def get_handler(self, _web_socket, _schema_id):
        """
        Find the appropriate handler for the schema

        :param _web_socket: If applicable, a web socket
        :param _schema_id: The message schema id
        """

        try:
            _category = schema_categories[_schema_id]
        except KeyError:
            # A proper running system should never encounter this error, so this is considered a concious probe
            self.handle_error("No category found for schema Id " + str(_schema_id),
                _web_socket=_web_socket, _category=EC_PROBE, _severity=SEV_ERROR)
            return None
        try:
            return self.category_shortcut[_category]
        except KeyError:
            raise Exception(self.log_prefix + "No handler for category: " + str(_category))

    def handle(self, _item):
        """
        The handle function is called when a monitor finds an unhandled item on its queue.
        The WebSocketHandler uses the schemaRef and schema categories to find the correct handler for the item.
        :param _item: The message
        """

        _web_socket = _item[0]
        _message_data = _item[1]

        if _web_socket:
            self.write_dbg_info("Handling " + _web_socket.address + " - message : " + str(_message_data))
        else:
            self.write_dbg_info("Handling outgoing message : " + str(_message_data))

        try:
            _schema_id = _message_data["schemaRef"]
        except KeyError:
            # A proper running system should never encounter this error, so this is considered a concious probe
            self.handle_error("No schema id found in message",
                _category=EC_PROBE, _severity=SEV_ERROR, _web_socket = _web_socket)
            return

        _handler = self.get_handler(_web_socket, _schema_id)

        try:
            _handler(_web_socket, _message_data)
        except Exception as e:
            self.handle_error("Error running handler " + str(_schema_id) + " Error: " + str(e),
                _category=EC_INTERNAL, _severity=SEV_ERROR, _web_socket = _web_socket)


    def shut_down(self, _user_id, _code=GOING_AWAY):
        """
        Shut down the handler
        :param _user_id: The user initiating the shut down
        """
        # Stop web sockets
        self.write_dbg_info("shut_down: Closing sockets..")
        for session in self.peers.values():
            if "web_socket" in session:
                self.write_dbg_info(session["address"])
                session["web_socket"].close(code=_code, reason="Shutting down")

        super(WebSocketHandler, self).shut_down(_user_id)

    def register_web_socket(self, _web_socket):
        """
        Register a web socket with the handler. Set its address field and message queue.

        :param _web_socket:

        """
        try:
            _session = self.peers[_web_socket.session_id]
        except KeyError:
            raise Exception(self.log_prefix + "Invalid session id:" + _web_socket.session_id)

        self.web_socket_lock.acquire()
        try:

            # Set the web socket of the session
            _session["web_socket"] = _web_socket
            # Set the address of the socket
            _web_socket.address = _session["address"]  # Perhaps the web socket should set this itself, dunno..
            _web_socket.message_queue = _session["queue"]
            _web_socket.process_id = self.process_id
            _web_socket.address_own = self.address
            # Add the peer to the address session dict
            if _web_socket.address in self.address__session:
                self.write_dbg_info("Register_web_socket: The " + "\"" + _web_socket.address +
                    "\" peer was already registered. Earlier failure to unregister/disconnect? Overwriting the registration.")

            self.address__session[_web_socket.address] = _web_socket.session_id
            self.write_dbg_info(str(_web_socket.address) + " registered, session data: " + str(_session))

        finally:
            # Unlock
            self.web_socket_lock.release()

    def unregister_web_socket(self, _web_socket):
        """
        Unregisters the websocket from the handler

        :param _web_socket:
        """
        self.web_socket_lock.acquire()
        try:
            if _web_socket.address in self.address__session:
                del self.address__session[_web_socket.address]
            else:
                # TODO: Should this be a problem?
                self.write_dbg_info("unregister_web_socket: The " + "\"" + str(
                        _web_socket.address) + "\" peer wasn't registered.")
            if "web_socket" not in self.peers[_web_socket.session_id]:
                self.write_dbg_info(
                    self.log_prefix + "ERROR: Failing to unregister web socket completely as there is no web socket, " +
                    "address " + str(_web_socket.address) + "session: " +
                    str(self.peers[_web_socket.session_id]))
            else:
                del self.peers[_web_socket.session_id]["web_socket"]
            self.write_dbg_info(self.log_prefix + "web_socket (address " + str(_web_socket.address) +
                  ", session_id " + str(_web_socket.session_id) + ") unregistered.")
        finally:
            # Unlock
            self.web_socket_lock.release()



"""
This is the central logging facility of the Optimal framework.
"""
import datetime

__author__ = 'Nicklas Borjesson'
import os


# Severity levels
SEV_DEBUG = 0  # Debugging message
SEV_INFO = 1  # Informational message
SEV_WARNING = 2  # A warning
SEV_ALERT = 3  # Action must be taken immidiately
SEV_USER = 4  # A user error or error that can be corrected by the user
SEV_ERROR = 5  # A problem but doesn't stop execution
SEV_FATAL = 6  # A problem that causes something to stop functioning

# EVENT CATEGORIES

EC_NOTIFICATION = 0  # A notification

# Errors
EC_INTERNAL = 1  # An internal error, likely a bug in the system
EC_INVALID = 2  # A validation error, some information failed to validate, invalid reference
EC_COMMUNICATION = 3  # A communications error
EC_SERVICE = 4 # A service level error, such as a failure to start or configuration errors
EC_RESOURCE = 5  # A resource error indicating a lack of memory, space or time/cpu or other resource
EC_RIGHT = 6  # A security related error, insufficient permissions or rights
EC_PERMISSION = 7  # A security related error, insufficient permissions or rights
EC_UNCATEGORIZED = 8  # Uncategorized error

# Node change categories
EC_ADD = 9  # Something was added
EC_REMOVE = 10  # Something was removed
EC_CHANGE = 11  # Something was changed

# Attack categories
EC_PROBE = 12  # The system consider itself being probed
EC_DOS = 13  # The system consider itself being under a denial-of-service attack
EC_BREAKIN = 14  # The system consider itself being broken in to

# Human representations

severity_identifiers = [
    "debug",
    "information",
    "warning",
    "user",
    "alert",
    "error",
    "fatal"
]

severity_descriptions = [
    "debugging message",
    "information message",
    "warning message",
    "user correctable error",
    "action must be taken immediately",
    "error",
    "a fatal error cause loss of functionality"
]

category_identifiers = [
    "notification",
    "internal",
    "invalid",
    "communication",
    "service",
    "resource",
    "right",
    "permission",
    "uncategorized",
    "add",
    "remove",
    "change",
    "probe",
    "dos",
    "breakin"
]

category_descriptions = [
    "notification message",
    "internal error, likely a bug in the system",
    "validation error, information/structure failed to validate, invalid reference",
    "error during communication",
    "service",
    "error indicating a lack of memory, space or time/cpu",
    "insufficient rights",
    "insufficient permissions",
    "uncategorized error",
    "node was added",
    "node was removed",
    "node was changed",
    "the system is being probed for vurnerabilities",
    "the system is under a denial-of-service attack",
    "someone is trying to break-in to the system"
]



"""The logging callback"""
callback = None
severity = SEV_WARNING

if os.name == "nt":
    import win32api
    import win32security
    import win32evtlog

    import win32con
    import win32evtlogutil

    ph=win32api.GetCurrentProcess()
    th = win32security.OpenProcessToken(ph,win32con.TOKEN_READ)
    my_sid = win32security.GetTokenInformation(th,win32security.TokenUser)[0]


    eventtype_severity = {
        SEV_DEBUG: win32evtlog.EVENTLOG_INFORMATION_TYPE,
        SEV_INFO: win32evtlog.EVENTLOG_INFORMATION_TYPE,
        SEV_WARNING: win32evtlog.EVENTLOG_WARNING_TYPE,
        SEV_ALERT: win32evtlog.EVENTLOG_INFORMATION_TYPE,
        SEV_USER: win32evtlog.EVENTLOG_ERROR_TYPE,
        SEV_ERROR: win32evtlog.EVENTLOG_ERROR_TYPE,
        SEV_FATAL: win32evtlog.EVENTLOG_ERROR_TYPE
    }

    def write_to_event_log(_data, _log_name,_category,  _severity):
        _strcategory = category_to_description(_category, "invalid category")
        _textual = [_strcategory[0] + _strcategory[1:] + "- severity: " +
            severity_to_identifier(_severity=_severity),
            _data
        ]
        win32evtlogutil.ReportEvent(_log_name, 1, eventType=_severity,
                                    strings = _textual,
                                    data="Raw\0Data".encode("ascii"), sid=my_sid)


def severity_to_identifier(_severity, _error):
    """Returns a matching severity identifiers"""
    if isinstance(_severity, int) and 0 <= _severity < len(severity_identifiers):
        return str(severity_identifiers[_severity])
    else:
        return _error + str(_severity)


def severity_to_description(_severity, _error):
    """Returns a matching severity description"""
    if isinstance(_severity, int) and 0 <= _severity < len(severity_descriptions):
        return str(severity_descriptions[_severity])
    else:
        return _error + str(_severity)


def category_to_identifier(_category, _error):
    """Returns a matching category identifiers"""
    if isinstance(_category, int) and 0 <= _category < len(category_identifiers):
        return str(category_identifiers[_category])
    else:
        return _error + str(_category)


def category_to_description(_category, _error):
    """Returns a matching category description"""
    if isinstance(_category, int) and 0 <= _category < len(category_descriptions):
        return str(category_descriptions[_category])
    else:
        return _error + str(_category)


def write_to_log(_data, _category=EC_NOTIFICATION, _severity=SEV_INFO, _process_id=None, _user_id=None,
                 _occurred_when=None, _address=None, _node_id=None, _uid=None, _pid=None):
    """
    Writes a message to the log using the current facility
    :param _data: The error message
    :param _category: The event category (defaults to CN_NOTIFICATION)
    :param _severity: The severity of the error (defaults to SEV_INFO)
    :param _process_id: The current process id (defaults to the current pid)
    :param _user_id: The Id of the user (defaults to the current username)
    :param _occurred_when: The time of occurrance (defaults to the current time)
    :param _address: The peer address
    :param _node_id: An Id for reference (like a node id)
    :param _uid: The system uid
    :param _pid: The system pid
    """
    global callback, severity
    if _severity < severity:
        return _data

    _occurred_when = _occurred_when if _occurred_when is not None else str(datetime.datetime.utcnow())
    _pid = _pid if _pid is not None else os.getpid()
    _uid = _uid if _uid is not None else os.getlogin()

    if callback is not None:
        callback(_data, _category, _severity, _process_id, _user_id, _occurred_when, _address, _node_id, _uid, _pid)
    else:
        print("Logging callback not set, print message:\n" + make_textual_log_message(_data, _category,
                                                                                      _severity, _process_id, _user_id,
                                                                                      _occurred_when, _address, _node_id,
                                                                                      _uid, _pid))

    return _data

def make_event(_data, _category=None, _severity=None, _process_id=None, _user_id=None,
                             _occurred_when=None, _address=None,_node_id=None, _uid=None, _pid=None):

    _event = (
        {
            "data": _data,
            "category": category_identifiers[_category],
            "severity": severity_identifiers[_severity],
            "uid": _uid,
            "pid": _pid,
            "occurredWhen": _occurred_when,
            "process_id": _process_id,
            "schemaRef": "of://event.json"
        }
    )
    if _node_id is not None:
        _event["node_id"] = _node_id
    if _user_id is not None:
        _event["user_id"] = _user_id
    if _address is not None:
        _event["address"] = _address
    if _process_id is not None:
        _event["process_id"] = _process_id

    return _event

def make_textual_log_message(_data, _category=None, _severity=None, _process_id=None, _user_id=None,
                             _occurred_when=None, _address=None,_node_id=None, _uid=None, _pid=None):
    """
    Build a nice textual error message based on available information for dialogs or event logs.

    :param _data: The message text or data
    :param _category: The kind of error
    :param _severity: The severity of the error
    :param _process_id: The current process id
    :param _user_id: The Id of the user
    :param _occurred_when: The time of occurrance
    :param _address: The peer address
    :param _node_id: An Id for reference (like a node id)
    :param _uid: The system uid
    :param _pid: The system pid

    :return: An error message

    """

    _result = "Process Id: " + (str(_process_id) if _process_id is not None else "Not available")
    _result += ", Adress: " + (str(_address) if _address is not None else str("None"))
    _result += (" - An error occurred:\n" if _severity > 2 else " - Message:\n") + str(_data)
    _result += "\nEvent category: " + category_to_identifier(_category,
                                                             "invalid event category:") if _category is not None else ""
    _result += "\nSeverity: " + severity_to_identifier(_severity,
                                                       "invalid severity level:") if _severity is not None else ""
    _result += "\nUser Id: " + str(_user_id) if _user_id is not None else ""
    _result += "\nOccurred when: " + str(_occurred_when) if _occurred_when is not None else ""
    _result += "\nEntity Id: " + str(_node_id) if _node_id is not None else ""
    _result += "\nSystem uid: " + (str(_uid) if _uid is not None else str(os.getlogin()))
    _result += "\nSystem pid: " + (str(_pid) if _pid is not None else str(os.getpid()))

    return _result


def make_sparse_log_message(_data, _category=None, _severity=None, _process_id=None, _user_id=None,
                             _occurred_when=None, _address=None, _node_id=None, _uid=None, _pid=None):
    """
    Build a sparse textual error message based on available information. One row unless data is multirow.

    :param _data: The message text or data
    :param _log_type: The type of data
    :param _category: The kind of error
    :param _severity: The severity of the error
    :param _process_id: The current process id
    :param _user_id: The Id of the user
    :param _occurred_when: The time of occurrance
    :param _address: The peer address
    :param _node_id: An Id for reference (like a node id)
    :param _uid: The system uid
    :param _pid: The system pid

    :return: An error message

    """
    if _severity == SEV_DEBUG:
        _prefix = " "
    elif _severity in [SEV_INFO, SEV_ALERT, SEV_WARNING]:
        _prefix = "-"
    else:
        _prefix = "*"

    _result = _prefix + "a: " + (str(_address) if _address is not None else str("None"))
    _result += ", pid: " + (str(_pid) if _pid is not None else str(os.getpid()))
    _result += ", uid: " + (str(_uid) if _uid is not None else str(os.getlogin()))

    if "\n" in _data:
        _result += ", data: (multirow, see below)"
        _result += ", ec: " + category_to_identifier(_category,"INV") if _category is not None else "N/A"
        _result += ", sev: " + severity_to_identifier(_severity,"INV") if _severity is not None else "N/A"
        _result += ", p_id: " + (str(_process_id) if _process_id is not None else "N/A")
        _result += ", u_id: " + str(_user_id) if _user_id is not None else ""
        _result += ", t: " + str(_occurred_when) if _occurred_when is not None else ""
        _result += ", node_id: " + str(_node_id) if _node_id is not None else ""
        _result += "\n===\n" + str(_data) +"\n=== "
    else:
        _result += ", data: " + str(_data) + ", "
        _result += "ec: " + category_to_identifier(_category,"INV") if _category is not None else "N/A"
        _result += ", sev: " + severity_to_identifier(_severity,"INV") if _severity is not None else "N/A"
        _result += ", p_id: " + (str(_process_id) if _process_id is not None else "N/A")
        _result += ", u_id: " + str(_user_id) if _user_id is not None else ""
        _result += ", t: " + str(_occurred_when) if _occurred_when is not None else ""
        _result += ", node_id: " + str(_node_id) if _node_id is not None else ""

    return _result



"""
This module holds the BrokerWebSocketHandler class
"""
from of.common.logging import write_to_log, EC_COMMUNICATION, SEV_INFO, SEV_DEBUG, EC_NOTIFICATION, SEV_ERROR
from of.common.messaging.constants import UNACCEPTABLE_DATA, BROKER_SHUTTING_DOWN
from of.common.messaging.handler import WebSocketHandler
from of.schemas.constants import intercept_schema_ids
from of.broker.globals import states, states_lookup

__author__ = 'Nicklas Borjesson'


class BrokerWebSocketHandler(WebSocketHandler):
    """
    This is the main handler class in the Optimal BPM system.
    It is called by the message queue monitor when an item appears on the queue.
    It is responsible for brokering messages between clients, writing their log message to the log
    and write their data to they database.
    """

    #: A DatabaseAccess instance for database connectivity
    database_access = None

    def __init__(self, _process_id, _peers, _schema_tools, _address, _database_access):
        """
        :param _process_id: The processId of the broker
        :param _peers: A dictionary of logged in peers
        :param _schema_tools: A SchemaTools instance for validation
        :param _address: The peer address of the broker
        :param _database_access: A DatabaseAccess instance
        """
        super(BrokerWebSocketHandler, self).__init__(_process_id, _peers, _schema_tools, _address)
        self.database_access = _database_access
        # Create a handler function map based on schema categories. See of.schemas.constants
        self.category_shortcut.update({
            "process": self.handle_process,
            "log": self.handle_logging,
            "message": self.handle_message,
            "control": self.handle_message
        })

    def handle_message(self, _source_web_socket, _message_data):
        """
        Handle a message
        :param _source_web_socket: The source web socket if external
        :param _message_data: The data in the message
        """
        if _source_web_socket is None:
            # Message came from internally, no need to validate.
            _message_data["source"] = self.address
        else:
            self.schema_tools.validate(_message_data)

        # Special case: A process result message should be intercepted and saved to the log.
        if _message_data["schemaRef"] in intercept_schema_ids:
            self.handle_logging(_source_web_socket, _message_data)

        _destination = _message_data["destination"]
        if _destination == self.address:
            # The reason for this is likely that something was instigated by the web client.
            write_to_log("Broker was the addressee, the broker is never a destination for a plain message.",
                         _category=EC_COMMUNICATION, _severity=SEV_ERROR, _process_id=self.process_id)
            return

        try:
            _destination_session = self.peers[self.address__session[_destination]]
        except KeyError:
            if _source_web_socket:
                _source_web_socket.close(code=UNACCEPTABLE_DATA,
                                         reason="Missing or invalid destination = " + _destination)
            else:
                raise KeyError("Missing or invalid destination = " + _destination)
        else:
            _destination_session["web_socket"].queue_message(_message_data)

    def handle_process(self, _web_socket, _process_data):
        """
        Writes incoming process information to the process collection.

        :param _web_socket: The source web socket
        :param _process_data: The data to write.
        """
        # Handle messages that writes to the backend, like process instances
        write_to_log("before saving process information, session: " + str(self.peers[_web_socket.session_id]),
                     _category=EC_NOTIFICATION, _severity=SEV_DEBUG, _process_id=self.process_id)

        self.database_access.save(_process_data, self.peers[_web_socket.session_id]["user"], _allow_save_id=True)

        write_to_log("handle_process succeeded", _category=EC_NOTIFICATION, _severity=SEV_DEBUG,
                     _process_id=self.process_id)

    def handle_logging(self, _web_socket, _log_data):
        """
        Write log items to the log collection

        :param _web_socket: The source web socket
        :param _log_data: The log data to write.
        """

        # Check integrity

        self.database_access.schema_tools.apply(_log_data)
        if _web_socket:
            # TODO: Should writtenBy be in some base log schema? (PROD-32)
            _log_data["writtenBy"] = self.peers[_web_socket.session_id]["user"]["_id"]
            _log_data["address"] = _web_socket.address

        self.database_access.logging.write_log(_log_data)

        write_to_log("handle_logging succeeded", _category=EC_NOTIFICATION, _severity=SEV_DEBUG,
                     _process_id=self.process_id)
        # Store states
        if _log_data["schemaRef"] == "of://log_process_state.json":
            # Handle the objectIds
            _log_data["_id"] = str(_log_data["_id"])
            _log_data["processId"] = str(_log_data["processId"])

            if _log_data["processId"] not in states_lookup:
                states.append(_log_data)
                states_lookup[_log_data["processId"]] = len(states) - 1
            else:
                states[states_lookup[_log_data["processId"]]] = _log_data

    def shut_down(self, _user_id):
        """
        Shut down the handler, also shuts down the worker and message monitors
        :param _user_id: The user_id that initiated the shut down
        :return:
        """
        super(BrokerWebSocketHandler, self).shut_down(_user_id, BROKER_SHUTTING_DOWN)



import os
import sys
import time
from multiprocessing import Process

import cherrypy
from bson.objectid import ObjectId
from pymongo.mongo_client import MongoClient

from ws4py.server.cherrypyserver import WebSocketPlugin, WebSocketTool

__author__ = "Nicklas Borjesson"

# The directory of the current file
script_dir = os.path.dirname(__file__)

# Add relative optimal bpm path to be able to load the modules of this repository properly
sys.path.append(os.path.join(script_dir, "../../"))

# IMPORTANT: ALL OPTIMAL FRAMEWORK IMPORTS MUST BE AFTER ADDING THE PATH
import of.common.logging
from of.broker.lib.access import DatabaseAccess
from of.broker.lib.auth_backend import MongoDBAuthBackend
from of.common.cumulative_dict import CumulativeDict
from of.common.logging import write_to_log, SEV_FATAL, EC_SERVICE, SEV_DEBUG, \
    EC_UNCATEGORIZED, SEV_ERROR, SEV_INFO, EC_INVALID, make_sparse_log_message, make_textual_log_message, make_event
from of.common.security.authentication import init_authentication
from of.common.settings import JSONXPath
from of.schemas.schema import SchemaTools
from of.broker import run_broker
from of.common.internal import register_signals, resolve_config_path
from of.common.messaging.factory import store_process_system_document, log_process_state_message
from of.broker.lib.messaging.websocket import BrokerWebSocket
from of.schemas.constants import zero_object_id
from of.schemas.validation import of_uri_handler
from of.broker.cherrypy_api.broker import CherryPyBroker
from of.broker.cherrypy_api.plugins import CherryPyPlugins

from of.broker.lib.messaging.handler import BrokerWebSocketHandler
from of.common.queue.monitor import Monitor
import of.common.messaging.websocket

if os.name == "nt":
    from of.common.logging import write_to_event_log

aux_runner = None

"""
Global variables
"""

#: The peer address of the broker
address = ""
#: A SchemaTools instance, used across the broker TODO: Is this thread/process safe?(ORG-112)
schema_tools = None
#: A DatabaseAccess instance, used across the broker TODO: Is this thread/process safe?(ORG-112)
database_access = None

#: The processId of the broker used to identify the process in logging
process_id = None
# Note: System pids are frequently reused on windows and may conflict on all platforms after reboots

#: The root web server class
web_root = None
#: The WebSocketPlugin instance
web_socket_plugin = None
# The web server configuration
web_config = None
# The plugins instance
plugins = None

# All namespaces
namespaces = None

# The severity when something is logged to the database
log_to_database_severity = None


def write_srvc_dbg(_data):
    global process_id
    write_to_log(_data, _category=EC_SERVICE, _severity=SEV_DEBUG, _process_id=process_id)


def log_locally(_data, _category, _severity, _process_id_param, _user_id, _occurred_when, _address_param, _node_id,
                _uid, _pid):
    global process_id, address

    if _process_id_param is None:
        _process_id_param = process_id
    if _address_param is None:
        _address_param = address
    if os.name == "nt":
        write_to_event_log(make_textual_log_message(_data, _data, _category, _severity, _process_id_param, _user_id,
                                                    _occurred_when, _address_param, _node_id, _uid, _pid),
                           "Application", _category, _severity)
    else:
        print(
            make_sparse_log_message(_data, _category, _severity, _process_id_param, _user_id, _occurred_when,
                                    _address_param, _node_id, _uid,
                                    _pid))
        # TODO: Add support for /var/log/message


def log_to_database(_data, _category, _severity, _process_id_param, _user_id, _occurred_when, _address_param, _node_id,
                    _uid, _pid):
    global log_to_database_severity, process_id, address

    if _process_id_param is None:
        _process_id_param = process_id
    if _address_param is None:
        _address_param = address

    if _severity < log_to_database_severity:
        log_locally(_data, _category, _severity, _process_id_param, _user_id, _occurred_when, _address_param,
                    _node_id, _uid, _pid)
    else:
        try:
            database_access.logging.write_log(make_event(_data, _category, _severity, _process_id_param, _user_id,
                                                         _occurred_when, _address_param, _node_id, _uid, _pid))
        except Exception as e:
            log_locally("Failed to write to database, error: " + str(e), EC_UNCATEGORIZED, SEV_ERROR,
                        _process_id_param, _user_id, _occurred_when, _address_param, _node_id, _uid, _pid)

        log_locally(_data, _category, _severity, process_id, _user_id, _occurred_when, _address_param, _node_id, _uid,
                    _pid)


def start_broker():
    """
    Starts the broker; Loads settings, connects to database, registers process and starts the web server.
    """

    global process_id, database_access, address, web_socket_plugin, repository_parent_folder, \
        web_config, schema_tools, namespaces, log_to_database_severity, plugins

    process_id = str(ObjectId())

    of.common.logging.callback = log_locally

    write_srvc_dbg("=====Starting broker=============================")

    try:
        _cfg_filename = resolve_config_path()
        _settings = JSONXPath(_cfg_filename)

    except Exception as e:
        if os.name == "nt":
            write_to_log(_data="Error loading settings from " + _cfg_filename,
                         _category=EC_SERVICE, _severity=SEV_FATAL)
        raise Exception("Error loading settings:" + str(e))

    of.common.logging.severity = of.common.logging.severity_identifiers.index(
        _settings.get("broker/logging/severityLevel", _default="warning"))

    log_to_database_severity = of.common.logging.severity_identifiers.index(
        _settings.get("broker/logging/databaseLevel", _default="warning"))

    write_srvc_dbg("Loaded settings from " + _cfg_filename)

    # An address is completely neccessary.
    address = _settings.get("broker/address", _default=None)
    if not address or address == "":
        write_to_log(_data="Broker cannot start, missing [broker] address setting in configuration file.",
                     _category=EC_SERVICE, _severity=SEV_FATAL)
        raise Exception("Broker cannot start, missing address.")

    # TODO: Reorganize. It is likely that almost everything but external database credentials should be stored in the db PROD-105

    # Initialize schema tools (of_uri_handler is later replaced by the general one)
    schema_tools = SchemaTools(_json_schema_folders=[os.path.join(script_dir, "../schemas/")],
                               _uri_handlers={"of": of_uri_handler})

    namespaces = CumulativeDict(_default={"schemas": []})

    write_srvc_dbg("Load plugin data")
    # Find the plugin directory
    _plugin_dir = _settings.get_path("broker/pluginFolder", _default="plugins")
    # Find the plugin directory
    repository_parent_folder = _settings.get_path("broker/repositoryFolder", _default="broker_repositories")
    # Load all plugin data
    plugins = CherryPyPlugins(_plugin_dir=_plugin_dir, _schema_tools=schema_tools, _namespaces=namespaces,
                               _process_id=process_id)

    # Plugins may want to initialize or add globals
    plugins.call_hook("init_globals", _broker_scope=globals())


    write_srvc_dbg("===Register signal handlers===")
    register_signals(stop_broker)
    plugins.call_hook("before_db_connect", _broker_scope=globals())
    # Connect to the database
    _host = _settings.get("broker/database/host", _default="127.0.0.1")
    _user = _settings.get("broker/database/username", _default=None)
    _password = _settings.get("broker/database/password", _default=None)
    if _user:
        write_srvc_dbg("===Connecting to remote MongoDB backend " + _host + "===")
        # http://api.mongodb.org/python/current/examples/authentication.html
        _client = MongoClient("mongodb://" + _user + ":" + _password + "@" + _host)
    else:
        write_srvc_dbg("===Connecting to local MongoDB backend===")
        _client = MongoClient()

    _database_name = _settings.get("broker/database/databaseName", _default="optimalframework")
    write_srvc_dbg("Using database name :" + _database_name)

    _database = _client[_database_name]
    database_access = DatabaseAccess(_database=_database, _schema_tools=schema_tools)
    of.common.logging.callback = log_to_database
    database_access.save(store_process_system_document(_process_id=process_id,
                                                       _name="Broker instance(" + address + ")"),
                         _user=None,
                         _allow_save_id=True)
    plugins.call_hook("after_db_connect", _broker_scope=globals())
    # TODO: It is possible that one would like to initialize, or at least read the plugins *before* trying to connect to the database

    # Must have a valid CherryPy version
    if hasattr(cherrypy.engine, "subscribe"):  # CherryPy >= 3.1
        pass
    else:
        write_to_log(_data="This application requires CherryPy >= 3.1 or higher.", _category=EC_SERVICE,
                     _severity=SEV_FATAL)
        raise Exception("Broker init: This application requires CherryPy >= 3.1 or higher.")
        # cherrypy.engine.on_stop_engine_list.append(_save_data)

    def ssl_path():
        # Figure out the path to the ssl-certificates
        # TODO: Load from database instead. Or not? (PROD-19)
        return os.path.join(os.path.expanduser("~"), "optimalframework")

    # Initialize CherryPy:s global configuration; note that this could be moved into a configuration file
    cherrypy.config.update({
        "tools.encode.on": True,
        "tools.encode.encoding": "utf-8",
        "tools.decode.on": True,
        "tools.trailing_slash.on": True,

        "tools.staticdir.root": os.path.abspath(os.path.dirname(__file__)),
        "server.ssl_module": "builtin",
        # TODO: Remove this when this bug is fixed:
        # https://bitbucket.org/cherrypy/cherrypy/issue/1341/autoreloader-also-fails-if-six-is-present
        "engine.autoreload.on": False,
        'server.socket_host': '0.0.0.0',
        "server.ssl_certificate": os.path.join(ssl_path(), "optimalframework_test_cert.pem"),
        "server.ssl_certificate": os.path.join(ssl_path(), "optimalframework_test_cert.pem"),
        "server.ssl_private_key": os.path.join(ssl_path(), "optimalframework_test_privkey.pem")
    })
    write_srvc_dbg("Starting CherryPy, ssl at " + os.path.join(ssl_path(), "optimalframework_test_privkey.pem"))

    web_config = {
        # The UI root
        "/": {
            "tools.staticdir.on": True,
            "tools.staticdir.dir": "ui",
            "tools.trailing_slash.on": True,
            "tools.staticdir.index": "index.html",
        },
        # Note that the web socket handling is put under /socket.
        "/socket": {
            "tools.websocket.on": True,
            "tools.websocket.handler_cls": BrokerWebSocket
        }
    }

    global web_root

    cherrypy._global_conf_alias.update(web_config)
    web_socket_plugin = WebSocketPlugin(cherrypy.engine)
    web_socket_plugin.subscribe()
    cherrypy.tools.websocket = WebSocketTool()

    cherrypy.engine.signals.bus.signal_handler.handlers = {'SIGUSR1': cherrypy.engine.signals.bus.graceful}

    # Initialize the decorator-based authentication framework
    init_authentication(MongoDBAuthBackend(database_access))

    # Initialize root UI
    web_root = CherryPyBroker(_process_id=process_id, _address=address, _database_access=database_access)
    # Initialize messaging
    of.common.messaging.websocket.monitor = Monitor(_handler=BrokerWebSocketHandler(process_id, _peers=web_root.peers,
                                                                                    _database_access=database_access,
                                                                                    _schema_tools=database_access.schema_tools,
                                                                                    _address=address))

    web_root.plugins = plugins
    # Generate the static content, initialisation
    plugins.call_hook("init_web", _broker_scope = globals())

    _web_config_debug = "Broker configured. Starting web server. Web config:\n"
    for _curr_key, _curr_config in web_config.items():
        if "tools.staticdir.dir" in _curr_config:
            _web_config_debug += "Path: " + _curr_key + " directory: " + _curr_config["tools.staticdir.dir"]
        else:
            _web_config_debug += "Path: " + _curr_key + " - no static dir"

    write_to_log(_web_config_debug, _category=EC_SERVICE, _severity=SEV_INFO)
    plugins.call_hook("pre_webserver_start", web_config=web_config, globals=globals())
    cherrypy.quickstart(web_root, "/", web_config)


def stop_broker(_reason, _restart=None):
    global web_root, process_id
    if _restart:
        write_to_log("BROKER WAS TOLD TO RESTART, shutting down orderly",
                     _category=EC_SERVICE, _severity=SEV_INFO, _process_id=process_id)
    else:
        write_to_log("BROKER WAS TERMINATED, shutting down orderly",
                     _category=EC_SERVICE, _severity=SEV_INFO, _process_id=process_id)

    write_srvc_dbg("Reason:" + str(_reason))

    _exit_status = 0
    write_srvc_dbg("Stop the monitor")
    try:
        of.common.messaging.websocket.monitor.stop()
    except Exception as e:
        write_to_log("Exception trying to stop monitor:", _category=EC_INVALID)
        _exit_status += 1
    time.sleep(1)

    # TODO: Terminate all child processes.(ORG-112)

    try:
        database_access.save(log_process_state_message(_changed_by=zero_object_id,
                                                       _state="killed",
                                                       _process_id=process_id,
                                                       _reason="Broker was terminated, reason: \"" +
                                                                _reason + "\", shutting down gracefully"),
                             _user=None)

    except Exception as e:
        write_to_log("Exception trying to write log item to Mongo DB backend:" + str(e), _category=EC_SERVICE,
                     _severity=SEV_ERROR)
        _exit_status += 1

    try:

        write_srvc_dbg("Unsubscribing the web socket plugin...")
        web_socket_plugin.unsubscribe()

        write_srvc_dbg("Stopping the web socket plugin...")
        web_socket_plugin.stop()

        write_srvc_dbg("Shutting down web server...")
        cherrypy.engine.stop()

        write_srvc_dbg("Web server shut down...")
    except Exception as e:
        write_to_log("Exception trying to shut down web server:" + str(e), _category=EC_SERVICE,
                     _severity=SEV_ERROR)
        _exit_status += 4

    if _restart:
        write_srvc_dbg("Broker was told to restart, so it now starts a new broker instance...")

        _broker_process = Process(target=run_broker, name="optimalframework_broker", daemon=False)
        _broker_process.start()
        _broker_process.join()

    write_srvc_dbg("Broker exiting with exit status " + str(_exit_status))

    if os.name != "nt":
        os._exit(_exit_status)
    else:
        cherrypy.engine.exit()
        return _exit_status
        # TODO: Add monitoring of processes and killing those not responding, log states to broker. (ORG-112)


if __name__ == "__main__":
    start_broker()



"""
The factory module contains a number of functions that return message structures.
Use these function instead of building messages manually to follow changes.
"""

import datetime
import os

if os.name != "nt":
    import pwd

import socket



__author__ = 'Nicklas Borjesson'


def get_current_login():
    if os.name == "nt":
        return os.getlogin()
    else:
        return pwd.getpwuid(os.geteuid()).pw_name



def reply_with_error_message(_runtime_instance, _message, _error_message):
    """
    Takes data from the incoming message and the instance replying to send an error message back to the sender.
    :param _runtime_instance: The calling class (to get the process_id from it)
    :param _message: The message to reply to(to get sender and messageId)
    :param _error_message: The error message to send
    """
    _struct = {
        "sourceProcessId": _runtime_instance.process_id,
        "messageId": _message["messageId"],
        "errorMessage": _error_message,
        "schemaRef": "of://message_error.json"
    }

    if "source" in _message or _message["source"] == "":
        _struct["destination"] = _message["source"]

    if "destination" in _message or _message["source"] == "":
        _struct["source"] = _message["destination"]
    if "sourceProcessId" in _message or _message["sourceProcessId"] == "":
        _struct["destinationProcessId"] = _message["sourceProcessId"]

    return _struct


def store_process_system_document(_process_id, _name, _parent_id=None):
    """
    Creates a process instance structure, automatically sets systemPid, spawnedBy, host and spawnedWhen
    """
    _struct = {
        "_id": _process_id,
        "systemPid": os.getpid(),
        "spawnedBy": get_current_login(),
        "name": _name,
        "host": socket.getfqdn(),
        "spawnedWhen": str(datetime.datetime.utcnow()),
        "schemaRef": "of://process_system.json"
    }
    if _parent_id:
        _struct["parent_id"] = _parent_id,

    return _struct


def log_process_state_message(_changed_by, _state, _process_id, _reason):
    """
    Created a general process state document
    """
    return {
        "changedBy": _changed_by,
        "changedWhen": str(datetime.datetime.utcnow()),
        "state": _state,
        "reason": _reason,
        "processId": _process_id,
        "schemaRef": "of://log_process_state.json"
    }




"""
This module serves as an extension for the JSON schema library.
It is made in a similar extension-vein as seep (https://github.com/Julian/Seep)
"""

from bson.objectid import ObjectId
# noinspection PyProtectedMember
from jsonschema import _utils
from jsonschema.exceptions import ValidationError
import jsonschema.validators


def mbe_object_id(value):
    """
    Converts a string value to a MongoDB object value.

    Instead of generating a new value, which is the default behaviour of ObjectId(), it returns None if provided None,
    :param value: A string value. Can be empty.
    :return: None or a corresponding ObjectId-instance.

    """
    if value is not None:
        return ObjectId(value)
    else:
        return None


class MongodbValidator():
    """
    The MongodbValidator class is a plug-in for JSON schema that converts objectId-formatted strings into
    actual ObjectId instances if the objectId property is present in the schema,

    """
    """The main validator class, a descendent of the jsonschema validator"""
    validator_class = None
    """The object setter is called to change the value"""
    object_setter = None

    def __init__(self, resolver=None):
        """Initializes MongodbValidator by initializing  and setting the resolver"""
        self.init_classes()
        self.resolver = resolver

    def init_classes(self):
        """Initialize all dynamically defined classes"""
        self.object_setter = jsonschema.validators.create(
            meta_schema={}, validators={"properties": self._set_object_ids},
        )

        # noinspection PyAttributeOutsideInit
        self.base_validator = jsonschema.validators.Draft4Validator
        self.validator_class = jsonschema.validators.extend(
            self.base_validator, {
                "properties": self._properties_with_objectids,
                "type": self._type_except_objectid,
            }
        )

    def apply(self, data, mongodb_schema):
        """
        Apply the MBE schema to the data.

        :param data: A mongo schema
        :param mongodb_schema: mongodb_schema: a mongodb_schema (JSON Schema with Seep properties)
        :return: the affected data

        """
        self.validator_class(schema=mongodb_schema, resolver=self.resolver).validate(data)
        return data

    def validate(self, data, mongodb_schema):
        """
        Validate the data with the MBE schema.

        :param data: A mongo schema
        :param mongodb_schema: mongodb_schema: a mongodb_schema (JSON Schema with Seep properties)
        :return: the affected data

        """
        self.base_validator(schema=mongodb_schema, resolver=self.resolver).validate(data)


    def check_schema(self, schema):
        """
        Check a schema, raises ValidationError if validation failed and RefResolutionError if it can't find a
        referenced file.

        :param schema: The schema to check.

        """
        # TODO: Check, should this really be like this?
        jsonschema.validators.Draft4Validator({}, resolver=self.resolver).check_schema(schema)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _type_except_objectid(self, validator, types, instance, schema):
        """
        Had to override the internal type checker to avoid it stopping at ObjectId.
        There is no other way for ObjectIds to be set in JSON data except for here.
        """
        types = _utils.ensure_list(types)

        if not any(validator.is_type(instance, _type) for _type in types):
            if not isinstance(instance, ObjectId):
                yield ValidationError(_utils.types_msg(instance, types))

    def _properties_with_objectids(self, validator, properties, instance, schema):
        """
        Responsible for looping properties.
        """
        for error in self.base_validator.VALIDATORS["properties"](
                validator, properties, instance, schema
        ):
            yield error
        subschemas = [(instance, schema)]
        while subschemas:
            subinstance, subschema = subschemas.pop()
            self.object_setter(subschema).validate(subinstance)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _set_object_ids(self, validator, properties, instance, schema):
        """
        Translates objectIds in the current scope.
        """
        for _property, subschema in properties.items():
            if _property in instance:
                if "objectId" in subschema:
                    # print("setting property - " + _property)
                    curr_instance = instance[_property]
                    if curr_instance is not None and curr_instance != "":
                        instance[_property] = mbe_object_id(curr_instance)

                elif "type" in subschema and subschema["type"] == "array":
                    if "objectId" in subschema["items"]:
                        # print("found array - " + _property)
                        for curr_idx in range(0, len(instance[_property])):
                            curr_item = instance[_property][curr_idx]
                            if type(curr_item) is str and curr_item is not None and curr_item != "":
                                # print("setting array item - " + curr_item)
                                instance[_property][curr_idx] = mbe_object_id(curr_item)





"""
This module holds Optimal Framework-specific JSON schema functionality
"""
import json
import os
from urllib.parse import urlparse

__author__ = 'Nicklas Borjesson'

script_dir = os.path.dirname(__file__)


def general_uri_handler(_uri, _folder):
    """
    This function looks up a JSON schema that matches the URL in the given folder
    :param _uri: The _uri to handle
    :return: The schema
    """
    # TODO: If there ever was a typical function to memoize, this would be it.(ORG-112)

    # Use urlparse to parse the file location from the URI
    _file_location = os.path.abspath(os.path.join(_folder, urlparse(_uri).netloc))

    # noinspection PyTypeChecker
    with open(_file_location, "r", encoding="utf-8") as _schema_file:
        _json = json.load(_schema_file)
    return _json

def of_uri_handler(_uri):
    """
    This function is given as call back to JSON schema tools to handle the of:// namespace references
    :param _uri:
    :return: The schema
    """
    return general_uri_handler(_uri, script_dir)


def of_schema_folder():
    return script_dir


"""
The logging module handles all the logging to database of OF.
"""

import datetime

from of.broker.lib.schema_mongodb import mbe_object_id

__author__ = "nibo"


# coding=utf-8
"""
A dictionary difference calculator
Originally posted as:
http://stackoverflow.com/questions/1165352/fast-comparison-between-two-python-dictionary/1165552#1165552
"""


class DictDiffer(object):
    """
    Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values

    """

    def __init__(self, _old, _new):
        """
        Compares two dicts

        :param _old: The old dict
        :param _new: The new dict

        """

        self.current_dict, self.past_dict = _new, _old
        self.current_keys, self.past_keys = [
            set(d.keys()) for d in (_new, _old)
        ]
        self.intersect = self.current_keys.intersection(self.past_keys)

    def added(self):
        """
        A list of added items

        """
        return self.current_keys - self.intersect

    def removed(self):
        """
        Returns a list of removed items

        """
        return self.past_keys - self.intersect

    def changed(self):
        """
        Returns a list of changed items

        """
        return set(o for o in self.intersect
                   if self.past_dict[o] != self.current_dict[o])

    def unchanged(self):
        """
        Returns a list of unchanged items

        """
        return set(o for o in self.intersect
                   if self.past_dict[o] == self.current_dict[o])


class Logging():
    """
    The logging class provides functionality to properly write to the "log" collection
    """
    database = None
    _log_collection = None

    def __init__(self, _database):
        """
        Initialize the logger, connects to a database

        :param _database:

        """
        self.database = _database
        self._log_collection = self.database["log"]

    @staticmethod
    def _compare_documents(_old, _new):
        # TODO: This must probably be using field xpaths or something. JSON XPaths might be useful
        # The problem is if a field is in a list of objects. Then fieldId will not be unique in the list of objects.
        # How about always using xpaths for a change? The problem is fieldIds in lists.

        _changes = []
        _differ = DictDiffer(_old, _new)
        for _property in _differ.added():
            _changes.append({"propertyId": _property, "before": None, "after": _new[_property]})

        for _property in _differ.removed():
            _changes.append({"propertyId": _property, "before": _old[_property], "after": None})

        for _property in _differ.changed():
            _changes.append({"propertyId": _property, "before": _old[_property], "after": _new[_property]})

        return _changes

    @staticmethod
    def _generate_node_event_skeleton(_user_id, _occurred_when, _node_id):
        """
        Generate a skeleton for a event item

        :param _user_id: The userId
        :param _occurred_when: When the event occured
        :param _node_id: The affected nodeId

        """
        # TODO: Decide  on how to make user ID reach the logger.
        return {
            "user_id": mbe_object_id(_user_id),
            "occurredWhen": _occurred_when,
            "node_id": mbe_object_id(_node_id)
        }

    def write_error(self, _event):

        pass



    def write_log(self, _event):
        """
        Writes an event to the database

        :param _event: The event data

        """
        _event["writtenWhen"] = str(datetime.datetime.utcnow())
        self._log_collection.save(_event)

    def log_security(self, _category, _message, _user_id, _node_id):
        """
        Creates an security event log item

        :param _category: type of event, can be rights, access or attack
        :param _message: Error message
        :param _user_id: The _id of the user. A string(not the ObjectId)
        :param _node_id: If applicable, the concerned node document
        :return:

        """
        _event = self._generate_node_event_skeleton(_user_id, str(datetime.datetime.utcnow()), _node_id)
        _event["category"] = _category
        _event["severity"] = "user"
        _event["data"] = _message
        self.write_log(_event)

    def log_save(self, _document, _user_id, _old_document=None):
        """
        Creates an event item based on the differences between the provided documents

        :param _document: The new version of the document
        :param _user_id: The _id of the user. A string(not the ObjectId)
        :param _old_document: The previous version

        """

        # Overwrite existing value?
        _event = self._generate_node_event_skeleton(_user_id, str(datetime.datetime.utcnow()),
                                               _document["_id"])
        if _old_document:
            # If so, compare and generate differences and save an "change" event to the "log" collection
            _event["change"] = self._compare_documents(_old=_old_document, _new=_document)
            _event["category"] = "change"

        else:
            # Save an "add" event to "log" collection
            _event["add"] = _document
            _event["category"] = "add"

        self.write_log(_event)

    def log_remove(self, _removed_document, _user_id):
        """
        Created a "remove" event. Called when a document has been removed.

        :param _removed_document: The removed document.
        :param _user_id: The _id of the user. A string(not the ObjectId)

        """

        _event = self._generate_node_event_skeleton(_user_id, str(datetime.datetime.utcnow()),
                                               _removed_document["_id"])
        _event["remove"] = _removed_document
        _event["category"] = "remove"
        self.write_log(_event)





"""
This module handles the plugin structure of the  the Optimal BPM Admin API as a web service through a CherryPy module
"""
import importlib
import json
import os
import time
from urllib.parse import urlparse
from uuid import UUID
import sys

import cherrypy

from of.broker.cherrypy_api.node import aop_check_session
from of.common.cumulative_dict import CumulativeDict
from of.common.logging import write_to_log, EC_NOTIFICATION, SEV_DEBUG, SEV_ERROR, EC_SERVICE, SEV_INFO

__author__ = 'Nicklas Borjesson'


class CherryPyPlugins(object):
    """
    The CherryPyPlugins class is a plugin for CherryPy to handle the Optimal BPM plug-ins
    """

    """List of plugins"""
    plugins = None

    """Last refresh"""
    last_refresh_time = None

    """ Schema tools instance"""
    schema_tools = None


    """Reference to the namespaces"""
    namespaces = None
    """The process id (not pid)"""
    process_id = None

    def __init__(self, _plugin_dir, _schema_tools, _namespaces, _process_id):
        # TODO: Create a dicts schema
        self.schema_tools = _schema_tools
        self.last_refresh_time = -31
        self.namespaces = _namespaces
        self.process_id = _process_id

        # Add the parent of plugins to sys path
        sys.path.append(os.path.join(_plugin_dir, ".."))
        self.refresh_plugins(_plugin_dir)

    def validate_uuid(self, _value):
        try:
            UUID(_value, version=4)
            return True
        except:
            return False

    def write_debug_info(self, _data):
        write_to_log(_data=_data, _category=EC_NOTIFICATION, _severity=SEV_DEBUG, _process_id=self.process_id)

    def call_hook(self, _hook_name, **kwargs):
        self.write_debug_info("Running hook " + _hook_name)
        for _curr_plugin_name, _curr_plugin in self.plugins.items():
            if ("failed" in _curr_plugin and _curr_plugin["failed"]):
                self.write_debug_info("Plugin " + _curr_plugin_name + " marked failed, will not call its hook.")
                continue
            if "hooks_instance" in _curr_plugin:
                    _hooks_instance = _curr_plugin["hooks_instance"]
                    if hasattr(_hooks_instance, _hook_name):
                        try:
                            self.write_debug_info("Calling " + _hook_name + " in " + _curr_plugin["description"])
                            getattr(_hooks_instance, _hook_name)(**kwargs)
                        except Exception as e:
                            write_to_log(
                                "An error occurred " + "Calling " + _hook_name + " in " + _curr_plugin_name + ":" + str(
                                    e),
                                _category=EC_SERVICE, _severity=SEV_ERROR)
                            if "failOnError" in _curr_plugin and _curr_plugin["failOnError"]:
                                write_to_log(
                                    "Setting " + _curr_plugin_name + " as Failed. No more hooks will be called for this plugin.",
                                    _category=EC_SERVICE, _severity=SEV_INFO)
                                _curr_plugin["failed"] = True
                            else:
                                write_to_log("Ignores error, this plugin will continue to attempt initialization.",
                                         _category=EC_SERVICE, _severity=SEV_INFO)

    def load_plugin(self, _plugins_dir, _plugin_name):
        # Load definitions.json

        _dirname = os.path.join(_plugins_dir, _plugin_name)
        # TODO: Here any check for licensing should be made to define if a plugin should be loaded.

        self.write_debug_info("Loading plugin: " + _dirname)

        _definitions_filename = os.path.join(_dirname, "definitions.json")
        if not os.path.exists(_definitions_filename):
            raise Exception(
                "Error loading plugin " + _dirname + ", definition file (" + _definitions_filename + ") missing.")

        with open(_definitions_filename) as _f_def:
            _definitions_data = json.load(_f_def)

        _plugin_data = _definitions_data["plugins"][_plugin_name]
        _plugin_data["baseDirectoryName"] = _dirname
        self.write_debug_info("Loading plugin in " + _plugin_data["description"])

        # Load schemas from /schema
        _schema_dir = os.path.join(_dirname, "schemas")
        if os.path.exists(_schema_dir):

            _schema_dir_list = os.listdir(_schema_dir)

            for _curr_file in _schema_dir_list:
                if os.path.splitext(_curr_file)[1] == ".json":
                    self.write_debug_info("Loading schema " + _curr_file)
                    with open(os.path.join(_schema_dir, _curr_file)) as _f_def:
                        _curr_schema = json.load(_f_def)
                    if "namespace" in _curr_schema:
                        # Initiate the namespace (so this row has effect, but should instead be a call. For readability)
                        self.namespaces[_curr_schema["namespace"]]
                        # Store schema among the unresolved ones
                        self._unresolved_schemas[_curr_schema["namespace"] + "://" + _curr_file] = _curr_schema
                    else:
                        self.write_debug_info("No namespace defined in " + _curr_file + ", ignoring.")
        else:
            self.write_debug_info("No schema folder, not loading schemas.")

        # Add server side stuff

        if "hooks_module" in _plugin_data:
            _hooks_modulename = _plugin_data["hooks_module"]

            try:
                _module = importlib.import_module("plugins." + _plugin_name + "." + _hooks_modulename)
                _plugin_data["hooks_instance"] = _module
            except Exception as e:
                write_to_log("An error occurred importing " + _hooks_modulename + " in " + _plugin_data[
                    "description"] + ":" + str(e),
                             _category=EC_SERVICE, _severity=SEV_ERROR)
                if "failOnError" in _plugin_data and _plugin_data["failOnError"]:
                    write_to_log(
                        "Setting " + _hooks_modulename + " as Failed. No hooks will be called for this plugin.",
                        _category=EC_SERVICE, _severity=SEV_INFO)
                    _plugin_data["failed"] = True
                else:
                    write_to_log("Ignores error, the plugin will continue to attempt initialization.",
                                 _category=EC_SERVICE, _severity=SEV_INFO)

        # Add definitions
        if "namespaces" in _definitions_data:
            self.namespaces.add_cumulatively(_definitions_data["namespaces"])
        self.plugins.add_cumulatively(_definitions_data["plugins"])
        return _plugin_data

    def refresh_plugins(self, _plugins_dir):
        # If < 30 seconds since last refresh (or some other principle)
        _curr_time = time.time()
        if self.last_refresh_time - _curr_time > 30:
            return

        # Reset unresolved schemas
        self._unresolved_schemas = {}

        # Find plugins directory

        if not os.path.exists(_plugins_dir):
            raise Exception("Plugin initialisation failed, no plugin directory where expected(" + _plugins_dir + ")")

        # Loop plugins
        _plugin_names = os.listdir(_plugins_dir)
        self.plugins = CumulativeDict()
        for _plugin_name in _plugin_names:
            if _plugin_name != "__pycache__" and os.path.isdir(os.path.join(_plugins_dir, _plugin_name)):
                self.load_plugin(_plugins_dir, _plugin_name)
                self.write_debug_info("Loaded plugin " + _plugin_name)

        # Manually add the optimal framework ("of") namespace
        self.namespaces["of"]["schemas"] = [_curr_ref for _curr_ref in self.schema_tools.json_schema_objects.keys()]
        self.schema_tools.resolver.handlers = {"of": self.uri_handler}

        # Add same resolver for all the rest of the namespaces (these resolvers will persist throughout the system)
        self.schema_tools.resolver.handlers.update(
            {_curr_namespace: self.uri_handler for _curr_namespace in self.namespaces})

        # Resolve all schemas

        for _curr_schema_key, _curr_schema in self._unresolved_schemas.items():
            _resolved_schema = self.schema_tools.resolveSchema(_curr_schema)
            self.namespaces[_resolved_schema["namespace"]]["schemas"].append(_curr_schema_key)
            self.schema_tools.json_schema_objects[_curr_schema_key] = _resolved_schema

        self.write_debug_info("Schemas in " + str(", ").join(
            [_curr_plugin["description"] for _curr_plugin in self.plugins.values()]) + " loaded and resolved:  " +
                              str.join(", ",
                                       ["\"" + _curr_schema["title"] + "\"" for _curr_schema in
                                        self._unresolved_schemas.values()]))

        # Remember refresh
        self.last_refresh_time = _curr_time

    def uri_handler(self, uri):
        """

        Handle all schema references, also for unresolved schemas

        :param uri: The uri to handle
        :return: The schema
        """
        if uri in self.schema_tools.json_schema_objects:
            return self.schema_tools.json_schema_objects[uri]
        else:
            return self._unresolved_schemas[uri]



"""
Created on Nov 6, 2012

@author: Nicklas Börjesson
@node: Utilities to help creating aspects in AOP.
"""
from inspect import getfullargspec


def alter_function_parameter_and_call(function_object, args, kwargs, name, value, err_if_not_set=None):
    """Changes or sets the value of an argument, if present, and calls the function.

    :param function_object: An instance of the called function
    :param args: A list of arguments
    :param kwargs: A dict of keyword arguments
    :param name: Name of the parameter that should be set
    :param value: Value to set that parameter to
    :param err_if_not_set: If is set, raise as error if the argument isn't present.
    :return: The return value of the called function

    """
    argspec = getfullargspec(function_object)

    if name in argspec.args:
        arg_idx = argspec.args.index(name)
        largs = list(args)
        largs.pop(arg_idx)
        largs.insert(arg_idx, value)
        new_args = tuple(largs)
        return function_object(*new_args, **kwargs)
    elif err_if_not_set is not None:
        raise Exception(err_if_not_set)

    return function_object(args, kwargs)







"""
Created on Apr 8, 2016

@author: Nicklas Boerjesson
"""

from setuptools import setup

import sys, os

from of import __release__

if __name__ == "__main__":
    setup(
        name="of",
        version=__release__,
        description="The Optimal Framework is the fastest path to building any modern multi-user system",
        author="Nicklas Boerjesson",
        author_email="nicklas_attheold_optimalbpm.se",
        maintainer="Nicklas Boerjesson",
        maintainer_email="nicklas_attheold_optimalbpm.se",
        install_requires=["pymongo", "jsonschema", "decorator", "requests", "cherrypy", "pycrypto", "ws4py"],
        long_description="""The Optimal Framework is a turnkey, Python-based development framework.\n
        It starts off with a running server, a MongoDB backend and all baseline features of any multiuser system.""",
        url="https://github.com/OptimalBPM/of",
        packages=["of", "of.broker", "of.broker.cherrypy_api", "of.broker.examples", "of.broker.examples.partial", "of.broker.testing", "of.broker.lib", "of.broker.lib.messaging", "of.broker.ui", "of.broker.ui", "of.common",
                  "of.common.messaging", "of.common.queue", "of.common.security", "of.schemas"],
        package_data = {
            # If any package contains *.md, *.json, *.txt or *.xml, *.sql files, include them:
            "": ["*.md", "*.json", "*.txt", "*.xml", "*.sql", "*.html"]
        },
        license="BSD")



"""
Created on Mar 6, 2015

@author: Nicklas Boerjesson
@note: The schema tools

"""
from copy import deepcopy
import os
import json


from jsonschema.validators import RefResolver
from jsonschema.exceptions import SchemaError



# strict-rfc3339


from of.broker.lib.schema_mongodb import MongodbValidator
from of.common.logging import EC_NOTIFICATION, SEV_INFO, SEV_DEBUG, write_to_log


class SchemaTools():
    """
        The schema tools class does all the handling, validation and transformation of schemas in the optimal framework.
    """
    # The json_schema_objects keeps the parsed and instantiated schema objects.
    json_schema_objects = None
    # json_schema_folder is the folder of where the json schemas are kept
    json_schema_folders = None

    # The resolve is responsible for resolving URI prefixes referencing other schemas,
    resolver = None
    # The URI-handlers are a dict of callbacks, used when resolving URI:s.
    uri_handlers = None

    # An instance of the mongodb JSON validator.
    mongodb_validator = None


    # noinspection PyDefaultArgument
    def __init__(self, _json_schema_folders=[], _uri_handlers=None):
        """
        Initiate the SchemaTools class

        :param _json_schema_folders: A list of folders where schema files are stored
        :param : _uri_handlers: A dict of uri_handlers, resolves a URI prefix to a actual schema.

        """

        if not _json_schema_folders:
            _json_schema_folders = []

        if not _uri_handlers:
            self.uri_handlers = {}
        else:
            self.uri_handlers = _uri_handlers

        self.resolver = RefResolver(base_uri="",
                                handlers=self.uri_handlers, referrer=None, cache_remote=True)

        self.mongodb_validator = MongodbValidator(resolver= self.resolver)

        self.json_schema_objects = {}

        # Load application specific schemas
        for _curr_folder in _json_schema_folders:
            self.load_schemas_from_directory(os.path.abspath(_curr_folder))

        # Resolve all the schemas
        for curr_schemaRef, curr_schema in self.json_schema_objects.items():
            self.json_schema_objects[curr_schemaRef] = self.resolveSchema(curr_schema)

        write_to_log("Schemas loaded and resolved: " +
                     str.join(", ",  ["\"" +_curr_schema["title"] + "\""  for _curr_schema in self.json_schema_objects.values()])
                     , _category=EC_NOTIFICATION, _severity=SEV_DEBUG)


    @staticmethod
    def check_schema_fields(_curr_schema_obj, _curr_file):
        """ Check so all mandatory fields are in the schema
        :param _curr_schema_obj: Schema to check
        :param _curr_file: File name use in error message

        """

        def raise_field_error(_collection):
            raise Exception("MongoBackend.load_schemas_from_directory: The \"" + _collection + "\"" +
                            " field is not in the schema-\"" + _curr_file + "\"")

        if "version" not in _curr_schema_obj:
            raise_field_error("version")
        if "namespace" not in _curr_schema_obj:
            raise_field_error("namespace")

    def load_schema_from_file(self, _file_name):
        """
        Loads a specified schema from a file, checks it and stores it in the schema cache.

        :param _file_name: The name of the schema file

        """
        try:
            _curr_file = open(_file_name, "r")
        except Exception as e:
            raise Exception("load_schema_from_file: Error loading \"" + _file_name +
                            "\": " + str(e))
        try:
            _json_schema_obj = json.load(_curr_file)

        except Exception as e:
            raise Exception("load_schema_from_file: Error parsing \"" + _file_name +
                            "\"" + str(e))

        _curr_file.close()

        try:
            self.check_schema_fields(_json_schema_obj, _file_name)
        except SchemaError as scherr:
            raise Exception("load_schema_from_file: SchemaError in " + _file_name + " at path:" + str(
                scherr.path) + "\nMessage:\n" + str(scherr.message))
        except Exception as e:
            raise Exception("load_schema_from_file: schema validation in " + _file_name + ", error :" + str(e))
        _schemaRef = _json_schema_obj["namespace"] + "://" + os.path.split(_file_name)[1]
        self.json_schema_objects[_schemaRef] = _json_schema_obj

    def load_schemas_from_directory(self, _schema_folder):
        """
        Load and validate all schemas in a folder, add to json_schema_objects

        :param _schema_folder: Where to look

        """
        _only_files = [f for f in os.listdir(_schema_folder) if
                       os.path.isfile(os.path.join(_schema_folder, f)) and f[-5:].lower() == ".json"]
        for _file in _only_files:
            self.load_schema_from_file(os.path.join(_schema_folder, _file))

    def apply(self, _data, _schema_ref=None):
        """
        Validate the JSON in _data against a JSON schema.

        :param _data: The JSON data to validate
        :param _schema_ref: If set, validate against the specified schema, and not the one in the data.
        :return: the schema object that was validated against.

        """
        if _schema_ref is not None:
            _json_schema_obj = self.json_schema_objects[_schema_ref]
        else:
            if "schemaRef" in _data:
                try:
                    _json_schema_obj = self.json_schema_objects[_data["schemaRef"]]
                except KeyError as e:
                    raise Exception("SchemaTools.apply, invalid schemaRef: " + _data["schemaRef"])
            else:
                raise Exception("SchemaTools.apply, data must have a schemaRef attribute")

        self.mongodb_validator.apply(_data, _json_schema_obj)
        return _data, _json_schema_obj

    def validate(self, _data, _schema_ref=None):
        """
        Validate the JSON in _data against a JSON schema.

        :param _data: The JSON data to validate
        :param _schema_ref: If set, validate against the specified schema, and not the one in the data.
        :return: the schema object that was validated against.

        """
        if _schema_ref is not None:
            _json_schema_obj = self.json_schema_objects[_schema_ref]
        else:
            _json_schema_obj = self.json_schema_objects[_data["schemaRef"]]

        self.mongodb_validator.validate(_data, _json_schema_obj)
        return _data, _json_schema_obj


    def non_base_type(self, _type):
        if _type not in ["array", "string", "integer", "object"]:
            return [_type]
        else:
            return []

    def resolveSchema(self, _schema):
        """
        Recursively resolve all I{$ref} JSON references in a JSON Schema.
        :param _schema: A L{dict} with a JSON Schema.
        :return: The resolved JSON Schema, a L{dict}.
        """
        _result = deepcopy(_schema)



        def local_resolve(_obj, _ref_history):
            """
            Recurse the JSON-tree and see where there are unresolved remote references.
            :param _obj: The node to resolve
            :param _ref_history: The previously resolved remote references, for cyclical check
            """

            if isinstance(_obj, list):
                # Loop any list
                for item in _obj:
                    local_resolve(item, _ref_history)
                return

            if isinstance(_obj, dict):

                if "$ref" in _obj:
                    _curr_ref = _obj["$ref"]

                    # Do not resolve local references
                    if _curr_ref[0] == "#":
                        return

                    # Check for cyclical references
                    if _curr_ref in _ref_history:
                        raise Exception("Error, cyclical remote reference: " + str(_curr_ref) + ":  Formers " + str(_ref_history))

                    with self.resolver.resolving(_curr_ref) as resolved:
                        # Resolve the resolved schema
                        local_resolve(resolved, _ref_history + [_curr_ref])
                        # Remove the reference
                        del _obj["$ref"]
                        # Add the resolved fragment to the schema
                        _obj.update(resolved)


                else:
                    # Loop all properties
                    for _key, _value in _obj.items():
                        if isinstance(_value, dict) and "type" in _value:
                            local_resolve(_value, _ref_history)

                        else:
                            local_resolve(_value, _ref_history)
        try:
            local_resolve(_result, [])
        except Exception as e:
            raise Exception("schemaTools.resolveSchema: Error resolving schema:" + str(e) + "Schema " + json.dumps(_schema, indent=4))

        # Make top allOf into properties
        if "allOf" in _result:
            _new_properties = {}
            for _curr_properties in _result["allOf"]:
                _new_properties.update(_curr_properties["properties"])

            _result["properties"] = _new_properties
            del _result["allOf"]

        _result["$schema"] = "http://json-schema.org/draft-04/schema#"

        try:
            self.mongodb_validator.check_schema(_result)
        except Exception as e:
            raise Exception("schemaTools.resolveSchema: error validating resolved schema:" + str(e) + "Schema " + json.dumps(_result, indent=4))

        return _result